//===========================================================================
//
// TBan
//
// Ted Drain.  tnt1of2-samurize@yahoo.com
//
// FYI: Much of the documentation is from the Samurize help file.  The
// original source is at www.samurize.com.
//
//===========================================================================
#include "TBan.h"
#include <cstring>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include "resource.h"

//===========================================================================
//
// STATICS
//
//===========================================================================

// Reader thread handle.  File scope so that DllMain can terminate the
// thread when the DLL is unloaded. The thread won't be created and
// started until the first meter is running.
static HANDLE s_readerThread = 0;

// GUI variables.  The samurize window is needed as a parent for the
// dialog box and error message popups.  The DLL handle is required to
// load the resource information when the dialog is created.
static HWND s_samurizeWindow = 0;
static HINSTANCE s_dllHandle = 0;

// Reader error variable.  Only used to point to constant strings so
// we don't have to worry about memory management.  If the background
// reader thread has an error, it will trip all the Samurize functions
// to return that error instead of the normal string.  Seemed like the
// best way to make the background thread errors available.
static char* s_readerError = 0;

// USB controller device handle.  Initialize to zero to indicate that
// no connection had been made to the TBAN.  File scope so that if the
// DLL can close the connection if it's active when the DLL is
// unloaded.
static FT_HANDLE s_usbDevice = 0;

// TBAN data array.  The background TBAN reader thread will read
// values into this array and the module functions will retrieve
// values directly out this array.
static const int DATA_LEN = 285;
static char s_data[DATA_LEN];

// After data is read, the computations are done to convert to maximum
// fan RPM's which are stored here.
static int s_maxFanRPM[4];

// The device ID for the TBAN.  Read from the registry after plugin
// configuration.  File scoped since both the reader thread and the
// set commands need it.
static int s_deviceId;

// DEBUGGING ONLY.  If set to true, instead of talking to the TBAN, we
// just fill the array w/ random numbers.  Used for debugging.  When
// set to true, no communication is ever attempted with the TBAN.
static const bool s_useRandom = false;

// The location in the registry for storing the configuration data.
static const char* s_tbanRegKey = "Software\\Serious Samurize\\Plugins\\TBAN";

// Private functions.
static DWORD WINAPI readerStart( LPVOID lpParameter );
static void computeMaxFanRPM( char* data );
static char* readRandom( char* data );
static int CALLBACK dialogCallback( HWND hDialog, UINT uMsg, WPARAM wParam,
                                    LPARAM lParam );
static void errorDialog( HWND parent, const char* funcName, int errorCode=0 );
static int loadSettings( HWND parent, int& deviceId, int& refreshSec );
static int saveSettings( HWND parent, int deviceId, int refreshSec );


//===========================================================================
//
// SAMURIZE API FUNCTIONS
//
//===========================================================================

//: Library initialization.
//
// The standard DLLMain function. You can use this function to do
// one-off initialization and de-initialization for your plugin using
// the DLL_PROCESS_ATTACH and DLL_PROCESS_DETACH reasons for call.
//
//= INPUT VARIABLES
//
//- hModule             Handle to the DLL module.  Needed by GUI dialogs
//                      to load resources from.
//- ul_reason_for_call  Indicates why the DLL entry-point function is
//                      being called.  Value will be DLL_PROCESS_ATTACH if
//                      the DLL is being loaded by Samurize (ie. when the
//                      first meter that uses the plugin is created) and
//                      DLL_PROCESS_DETACH if DLL is being unloaded by
//                      Samurize (ie. after the last meter that uses the
//                      plugin has been destroyed)
//- lpReserved          If fdwReason is DLL_PROCESS_ATTACH, lpvReserved is
//                      NULL for dynamic loads and non-NULL for static loads.
//                      If fdwReason is DLL_PROCESS_DETACH, lpvReserved is
//                      NULL if DllMain has been called by using FreeLibrary
//                      and non-NULL if DllMain has been called during process
//                      termination.
//= RETURN VALUE
//
//- This function should return TRUE.
//
BOOL APIENTRY
DllMain( HINSTANCE hModule,
         DWORD ul_reason_for_call,
         LPVOID lpReserved )
{
   switch ( ul_reason_for_call )
   {
   // This is when the plugin is loaded
   case DLL_PROCESS_ATTACH:
      // Save a handle to the DLL
      s_dllHandle = hModule;
      
      // Zero out all the module variables.
      s_usbDevice = 0;
      s_readerError = 0;
      memset( &s_data[0], 0, DATA_LEN );
      break;

   // This is when the plugin is shut down
   case DLL_PROCESS_DETACH:
      if ( s_usbDevice )
      {
         // Shut down the TBAN connection.
         FT_STATUS status = FT_Close( s_usbDevice );
         if ( status != FT_OK )
         {
            // Not much we can do if this doesn't work so just ignore
            // the failure flag.
         }
         s_usbDevice = 0;
      }

      if ( s_readerThread )
      {
         TerminateThread( s_readerThread, 0 );
         CloseHandle( s_readerThread );
         s_readerThread = 0;
      }

      break;
   }

   return TRUE;
}

//: Get the list of functions in the module.
//
// The init() function tells Samurize the names of the functions your
// source plugin provides. It should return a list like so:
// "Func1|Func2|...|FuncN"
//
//= RETURN VALUE
//
//- A list of all the functions your source plugin provides eg. "My
//  Function 1|Some other function|And another"
//
SAMURIZE_PLUGIN_CHAR
init()
{   
   return "GetAnalogTemp|"
          "GetDigitalTemp|"
          "GetFanRPM|"
          "GetFanPercentage|"
          "GetFanMaxRPM|"
          "GetFanManualMode|"
          "GetValue";
   //"SetFanPercentage";
}


//: Tell Samurize what type of plugin this is.
//
// If this function is declared, Samurize will assume you are using
// the SDK v2.0.
//
// All source functions your plugin provides will be called with an
// additional first parameter, the meter's unique_id.  dllstartup will
// be called for each meter created that uses your plugin (if you
// provide that function).  dllshutdown when a meter that uses your
// plugin is deleted (if you provide that function)
//
//= RETURN VALUE
//
//- The value you should return is a summation of any or all of the
//  following values:  WM_SAM_VISUALPLUGIN, WM_SAM_INPUTPLUGIN,
//  WM_SAM_SOURCEPLUGIN.
//
SAMURIZE_PLUGIN_INT
dlltype()
{
   // These can be added together for more than one type.
   // WM_SAM_VISUALPLUGIN
   // WM_SAM_INPUTPLUGIN
   return WM_SAM_SOURCEPLUGIN;
}


//: Get general plugin information.
//
// This function should return information about your plugin depending
// on the passed integer value. This information is displayed when the
// 'About' button is pressed in the config editor when your plugin is
// selected.
//
// This function is optional. The input value is an enumeration for
// requesting the following values:
//
// 1: Plugin name 
// 2: Author name 
// 3: Plugin version 
// 4: Creation date 
// 5: Last modified date 
// 6: Website 
// 7: Contact email address 
//
//= RETURN VALUE
//
//- Return the requested information string.
//
SAMURIZE_PLUGIN_CHAR
getinfo( int infoNum )
{
   InfoType infoType = (InfoType)infoNum;
   switch ( infoType )
   {
   case PLUGIN_NAME: return "TBan";
   case PLUGIN_AUTHOR: return "Ted Drain";
   case PLUGIN_VERSION: return "1.1";
   case PLUGIN_CREATED: return "10-Feb-2006";
   case PLUGIN_UPDATED: return "15-Mar-2006";
   case PLUGIN_WEBSITE: return "None";
   case PLUGIN_EMAIL: return "tnt1of2-samurize@yahoo.com";
   }

   return "";
}

//: Meter startup function.
//
// dllstartup() is called every time Samurize creates a meter that
// uses this plugin.  Any memory allocation needed on a meter-by-meter
// basis should be done in this function.
//
//- INPUT VARIABLES
//
//- hwnd       The window handle of the instance of Samurize that called
//             this function. 
//- dlltype    The meter 'type' (1 for source plugins, 2 for visual plugins)
//             that is being initialized.
//
//= RETURN VALUE
//
//- This function should return a unique integer ID that Samurize
//  remembers and uses for other functions (this is so multiple
//  instances of Samurize that use this plugin do not conflict). If your
//  plugin is very simple and does not need to keep any global variables
//  that may cause conflicts, just return 0 (or omit this function).
//
SAMURIZE_PLUGIN_INT
dllstartup( HWND hwnd,
            int dlltype )
{
   // Save the window handle so we can use to make a GUI window.
   s_samurizeWindow = hwnd;
   
   // Start the reader thread.  Repeated calls to this are OK - it
   // won't do anything if the thread is already running.
   if ( ! s_readerThread )
   {
      // Create a thread for the reading of the TBAN data.  The thread
      // won't actually be started until a meter is started.
      s_readerThread = CreateThread( NULL,            // security attribute
                                     NULL,            // stack size
                                     readerStart,     // starting function
                                     NULL,            // function argument
                                     NULL,            // start now
                                     NULL );          // thread ID pointer
   }

   // Fixed length buffer for returning char* results.  Fixed length
   // is OK since the TBan data stream is a set of single byte values
   // that get converted to a string.
   char* buffer = new char[40];
   return (int)buffer;
}

//: Meter shutdown function.
//
// dllshutdown() is called every time a meter using the plugin is
// destroyed by Samurize.  Any memory releasing needed on a
// meter-by-meter basis should be done in this function.  If your
// plugin does not need to free any memory, this function may be
// omitted.
//
//= INPUT VARIABLES
//
//- id      The unique ID your plugin assigned to the meter
//          (returned by dllstartup)
//
//= RETURN VALUE
//
//- This function should always return 0.
//
SAMURIZE_PLUGIN_INT
dllshutdown( int id )
{
   // Leave the TBAN connected since there could be multiple modules
   // using it.
   if ( id )
   {
      char* buffer = (char*)id;
      delete [] buffer;
   }

   return 0;
}

//: Copy a meter.
//
// pastemeter is called when a meter is pasted in the config
// editor. Your plugin should clone the settings from the source meter
// (indicated by source_id) to the cloned meter (indicated by
// dest_id).
//
//= INPUT VARIABLES
//
//- source_id     The unique ID assigned to the source meter by the
//                dllstartup() function.
//- dest_id       The unique ID assigned to the destination meter by
//                the dllstartup() function
//
//= RETURN VALUE
//
//- This function should return 0.
//
SAMURIZE_PLUGIN_INT
pastemeter( int thesource,
            int thedest )
{
   // Do nothing.  Nothing to copy since we're just using the struct
   // as string buffer.
   return 0;
}

//: Get the required parameters for a module function.
//
// The getparam() function should return a list of the parameters that
// the passed function requires.
//
//= INPUT VARIABLES
//
//- func    The name of the function Samurize needs to know the parameters for.
//
//= RETURN VALUE
//
//- A list of parameters delimited by '|' characters. Note that there
//  needs to be a final '|' after the last parameter. Appending '=' and
//  then a value to the end of a parameter indicates the default value
//  for that parameter.  eg. "Param1|Param2|Param3=SomeDefaultValue|"
//
SAMURIZE_PLUGIN_CHAR
getparam( char* func )
{
   if ( strcmp( func, "GetDigitalTemp" ) == 0 )
   {
      return "Digital Sensor Index (0-7)|";
   }
   else if ( strcmp( func, "GetAnalogTemp" ) == 0 )
   {
      return "Analog Sensor Index (0-5)|";
   }
   else if ( strcmp( func, "GetFanRPM" ) == 0 )
   {
      return "Fan Index (0-3)|";
   }
   else if ( strcmp( func, "GetFanPercentage" ) == 0 )
   {
      return "Fan Index (0-3)|";
   }
   else if ( strcmp( func, "GetFanMaxRPM" ) == 0 )
   {
      return "Fan Index (0-3)|";
   }
   else if ( strcmp( func, "GetFanManualMode" ) == 0 )
   {
      return "Fan Index (0-3)|";
   }
   else if ( strcmp( func, "GetValue" ) == 0 )
   {
      return "Parameter ID (1-285)|";
   }
   /***********
   else if ( strcmp( func, "SetFanManualMode" ) == 0 )
   {
      return "Fan Index (0-3)|Manual Mode (0,1)";
   }
   else if ( strcmp( func, "SetFanPercentage" ) == 0 )
   {
      return "Fan Index (0-3)|Percentage (0-100)";
   }
   ************/

   return "|";
}

//: Global plugin configuration.
//
// configure() is called when the user clicks on the "Configure
// Plugin..." button in Samurize.  You can display a popup dialog with
// global settings for your plugin (eg. a box prompting for your
// location for a weather plugin). These settings will be applied for
// all meters that use this source plugin.
//
// Note that it is your responsibility to save these settings to the
// registry/settings file/whatever - they aren't stored by Samurize
// anywhere.
//
//= RETURN VALUE
//
//- This function should always return 0.
//
SAMURIZE_PLUGIN_INT
configure()
{
   // Create a modal dialog.  The first variable must be the DLL
   // instance or none of the resources will be found.  The actual
   // work for everything is done in the callback function.  This runs
   // the dialog and waits for the user to finish with it.
   int ret = DialogBox( s_dllHandle,                    // dll instance
                        MAKEINTRESOURCE( IDD_DIALOG ),  // the resource
                        s_samurizeWindow,               // parent window
                        dialogCallback );               // callback function
   if ( ret == -1 )
   {
      errorDialog( s_samurizeWindow, "CreateDialog" );
   }
   
   return 0;
}

//: Individual meter configuration.
//
// source_configure() is called when the user clicks on the
// "Configure..." button in Samurize.  You can display some kind of
// popup window that allows users to specify various display settings
// for this source plugin. Note that you need to store the settings
// for each meter individually for saving later on.
//
//= INPUT VARIABLES
//
//- id         The unique ID of the meter being configured (that your
//             plugin assigned to the meter as the return value of dllstartup) 
//- func       The name of the function this meter calls from the plugin .
//
//
//= RETURN VALUE
//
//- This function should always return 0.
//
SAMURIZE_PLUGIN_INT
source_configure( int id,
                  char* func )
{
   // There is no per-meter configuration data.  In this case, just do
   // a global configuration since people are very likely to just
   // click the configure button instead of the global configure
   // button.
   return configure();
}

//===========================================================================
//
// LIBRARY COMMAND FUNCTIONS
//
//===========================================================================

//: Get the temperature from a digital meter.
//
//= INPUT VARIABLES
//
//- id                   The Samurize plugin ID.  This is ID from dllstartup()
//- digitalSensorIndex   The digital sensor index to read.  Valid range
//                       is from [0,7].
//
SAMURIZE_PLUGIN_CHAR
GetDigitalTemp( int id,
                char* digitalSensorIndex )
{
   int index = atoi( digitalSensorIndex );
   if ( index < 0 || index > 7 )
   {
      return "Invalid sensor index";
   }
   else if ( s_readerError )
   {
      return s_readerError;
   }

   char* buffer = (char*)id;

   // Calibrated digital temperature readings start at ID 239 (1
   // based) so subtract one to get a 0 based index.  Temperatures are
   // reported as numbers of .5 degrees C.
   const int baseIndex = 239 - 1;
   double value = (double)( (int)s_data[ baseIndex + index ] / 2.0 );
   sprintf( buffer, "%4.1f", value );
   return buffer;
}


//: Get the temperature from an analog meter.
//
//= INPUT VARIABLES
//
//- id                   The Samurize plugin ID.  This is ID from dllstartup()
//- analogSensorIndex    The analog sensor index to read.  Valid range
//                       is from [0,5].
//
SAMURIZE_PLUGIN_CHAR
GetAnalogTemp( int id,
               char* analogSensorIndex )
{
   int index = atoi( analogSensorIndex );
   if ( index < 0 || index > 5 )
   {
      return "Invalid sensor index";
   }
   else if ( s_readerError )
   {
      return s_readerError;
   }

   char* buffer = (char*)id;

   // Calibrated digital temperature readings start at ID 247 (1
   // based) so subtract one to get a 0 based index.  Temperatures are
   // reported as numbers of .5 degrees C.
   const int baseIndex = 247 - 1;
   double value = (double)( (int)s_data[ baseIndex + index ] / 2.0 );
   sprintf( buffer, "%4.1f", value );
   return buffer;
}

//: Read the fan RPM value.
//
// This is a computed quantity.  The TBAN returns the fan setting as a
// percentage of the maximum value.
//
//= INPUT VARIABLES
//
//- id                   The Samurize plugin ID.  This is ID from dllstartup()
//- fanIndex             The fan index to read.  Valid range is from [0,3].
//
SAMURIZE_PLUGIN_CHAR
GetFanRPM( int id,
           char* fanIndex )
{
   int index = atoi( fanIndex );
   if ( index < 0 || index > 3 )
   {
      return "Invalid fan index";
   }
   else if ( s_readerError )
   {
      return s_readerError;
   }

   char* buffer = (char*)id;

   // Fan PWM values start at ID 138 (1 based) so subtract one to get
   // a 0 based index.
   const int baseIndex = 138 - 1;

   // The PWM value is 1/2 the percentage of the maximum from (so the
   // read PWM will be from 0->50).
   int pwmValue = 2 * s_data[ baseIndex + index ];
   int rpm = (int)( pwmValue / 100.0 * s_maxFanRPM[index] );
   
   sprintf( buffer, "%d", rpm );
   return buffer;
}

//: Get the current fan percentage.
//
// The return value will be the fan percentage as an integer from
// [0,100].
//
//= INPUT VARIABLES
//
//- id                   The Samurize plugin ID.  This is ID from dllstartup()
//- fanIndex             The fan index to read.  Valid range is from [0,3].
//
SAMURIZE_PLUGIN_CHAR
GetFanPercentage( int id,
                  char* fanIndex )
{
   int index = atoi( fanIndex );
   if ( index < 0 || index > 3 )
   {
      return "Invalid fan index";
   }
   else if ( s_readerError )
   {
      return s_readerError;
   }

   char* buffer = (char*)id;

   // Fan PWM values start at ID 138 (1 based) so subtract one to get
   // a 0 based index.
   const int baseIndex = 138 - 1;

   // The PWM value is 1/2 the percentage of the maximum from (so the
   // read PWM will be from 0->50).
   int pwmValue = 2 * s_data[ baseIndex + index ];
   sprintf( buffer, "%d", pwmValue );
   return buffer;
}

//: Get the fan manual mode flag.
//
//= INPUT VARIABLES
//
//- id                   The Samurize plugin ID.  This is ID from dllstartup()
//- fanIndex             The fan index to read.  Valid range is from [0,3].
//
//= RETURN VALUE
//
//- Returns 1 if the fan is in manual mode.  0 if it is not.
//
SAMURIZE_PLUGIN_CHAR
GetFanManualMode( int id,
                  char* fanIndex )
{
   int index = atoi( fanIndex );
   if ( index < 0 || index > 3 )
   {
      return "Invalid fan index";
   }
   else if ( s_readerError )
   {
      return s_readerError;
   }

   char* buffer = (char*)id;

   // Fan manual mode flag values start at ID 102 (1 based) so
   // subtract one to get a 0 based index.
   const int baseIndex = 102 - 1;
   sprintf( buffer, "%d", s_data[ baseIndex + index ] );
   return buffer;
}

//: Get the maximum fan RPM.
//
//= INPUT VARIABLES
//
//- id                   The Samurize plugin ID.  This is ID from dllstartup()
//- fanIndex             The fan index to read.  Valid range is from [0,3].
//
SAMURIZE_PLUGIN_CHAR
GetFanMaxRPM( int id,
              char* fanIndex )
{
   int index = atoi( fanIndex );
   if ( index < 0 || index > 3 )
   {
      return "Invalid fan index";
   }
   else if ( s_readerError )
   {
      return s_readerError;
   }

   char* buffer = (char*)id;
   sprintf( buffer, "%d", s_maxFanRPM[index] );
   return buffer;
}

//: Get an arbitrary value.
//
// This will return any value from the TBAN array.  See the protocol
// spreadsheet from http://www.mcubed-tech.com/eng/support_kit.htm.
//
//= INPUT VARIABLES
//
//- id              The Samurize plugin ID.  This is ID from dllstartup()
//- parameterID     The parameter to read.  These are from the ID column
//                  of the spreadsheet.  Valid range is from [1,285].
//
SAMURIZE_PLUGIN_CHAR
GetValue( int id,
          char* parameterID )
{
   int index = atoi( parameterID );
   if ( index < 1 || index > DATA_LEN )
   {
      return "Invalid value index";
   }
   else if ( s_readerError )
   {
      return s_readerError;
   }

   // Data ranges are from 1->185 so subtract one to get a 0 based
   // index.
   char* buffer = (char*)id;
   sprintf( buffer, "%d", (int)s_data[ index - 1 ] );
   return buffer;
}

//: Set the current fan percentage.
//
// This will set the fan to the requested percentage and toggle the
// manual mode flag to on.  
//
//= INPUT VARIABLES
//
//- id                   The Samurize plugin ID.  This is ID from dllstartup()
//- fanIndex             The fan index to read.  Valid range is from [0,3].
//- percentage           The percentage to set the fan to.  Valid inputs are
//                       0 to 100.
//
SAMURIZE_PLUGIN_CHAR
SetFanPercentage( int id,
                  char* fanIndex,
                  char* percentage )
{
   //FYI: Not ready yet
   int index = atoi( fanIndex );
   int percent = atoi( percentage );
   if ( index < 0 || index > 3 )
   {
      return "Invalid fan index";
   }
   else if ( percent < 0 || percent > 100 )
   {
      return "Invalid percentage";
   }
   else if ( s_readerError )
   {
      return s_readerError;
   }

   char sendBuf[2];
   switch ( index )
   {
   case 0: sendBuf[0] = 0x11; break;
   case 1: sendBuf[0] = 0x12; break;
   case 2: sendBuf[0] = 0x13; break;
   case 3: sendBuf[0] = 0x14; break;
   }

   if ( s_usbDevice )
   {
      return 0;
   }
   
   sendBuf[1] = (char)percent;

   FT_HANDLE usbDevice = 0;
   char* error = openConnection( s_deviceId, &usbDevice );
   if ( error )
   {
      return error;
   }

   DWORD numWritten;
   FT_STATUS status = FT_Write( usbDevice, &sendBuf[0], 2, &numWritten);
   if ( status != FT_OK || numWritten != 2 )
   {
      return "FT_Write: Can't send SER_SET command.";
   }

   return 0;
}


//===========================================================================
//
// READER THREAD FUNCTIONS
//
//===========================================================================

//: Start the reader thread.
//
// This is the starting function for the reader thread.  It will
// create a timer, read from the TBAN, and repeat (i.e. the thread
// will never end).  The thread will only end if an error message is
// encountered.  In that case, the static variable s_readerError will
// point to the error message.
//
//= INPUT VARIABLES
//
//- lpParameter      A parameter passed to the thread creation function.
//                   Unused.
//
//= RETURN VALUE
//
//- Returns a positive integer if there was an error.
//
static DWORD WINAPI 
readerStart( LPVOID lpParameter )
{
   // Read the settings from the registry so we can talk to the TBAN.
   int refreshSec = 0;
   if ( loadSettings( s_samurizeWindow, s_deviceId, refreshSec ) )
   {
      return 1;
   }
   
   HANDLE timer = CreateWaitableTimer( NULL,    // security attribute
                                       true,    // manual timer
                                       NULL );  // timer name
   if ( timer == NULL )
   {
      s_readerError = "Can't create timer.";
      return 2;
   }

   // FYI: a Windows file time is a count in units of 100 nanseconds.
   // To convert from seconds to ftime, we do:
   //    ftime = sec * ftime/sec
   //    ftime = sec * ftime/nanosec * nanosec/sec
   //    ftime = sec * .01 * 1e9
   //    ftime = 1e7 * sec
   const INT64 secToFtime = 10000000;

   // Start running at the requested time.  Negative means relative
   // time.
   LARGE_INTEGER dueTime;
   dueTime.QuadPart = -refreshSec * secToFtime;

   // Period is in millisec.  Tried to set this repeat every n
   // seconds.  For whatever reason, this isn't working.  The timer is
   // not being reset so just use a manual timer that we reset after
   // each call.
   //const LONG period = refreshSec * 1000;
   const LONG period = 0;
   
   if ( ! SetWaitableTimer( timer,      // the timer
                            &dueTime,   // start running here
                            period,     // repeat period
                            NULL,       // timer function
                            NULL,       // argument to timer function
                            false ) )   // power restore flag
   {
      s_readerError = "Can't start timer.";
      return 1;
   }

   // Loop infinitely.  The timer will trip every period, we'll make
   // the read and repeast.
   while( true )
   {
      // Wait for the timer to finish.
      if ( WaitForSingleObject( timer, INFINITE ) == WAIT_FAILED )
      {
         s_readerError = "Error in WaitForSingleObject.";
         break;
      }

      // Either fill with random numbers or read from the TBAN.
      if ( s_useRandom )
      {
         s_readerError = readRandom( &s_data[0] );
      }
      else
      {
         // Open the connection to the TBAN.  
         s_readerError = openConnection( s_deviceId, &s_usbDevice );
         if ( ! s_readerError )
         {
            // Read the data and close the connection.
            s_readerError = readFromTBan( s_usbDevice, &s_data[0] );
            FT_Close( s_usbDevice );
            s_usbDevice = 0;
         }
      }

      // NOTE: If an error occurs, wait until the next loop and try
      // again.

      // Reset the timer for the next run.
      if ( ! SetWaitableTimer( timer, &dueTime, period, NULL, NULL, false ) )
      {
         s_readerError = "Can't reset timer.";
         break;
      }
   }

   // Should never reach here.  If we do there was an error in the
   // timer loop above.
   CloseHandle( timer );
   CloseHandle( s_readerThread );
   s_readerThread = 0;
   return 1;
}

//===========================================================================
//
// LIBRARY API FUNCTIONS
//
//===========================================================================

//: Open the connection to the TBAN.
//
//= INPUT VARIABLES
//
//- deviceId      The USB device ID number of the TBAN.  The number can be
//                found by running the TBAN software and looking under
//                Options/USB connections and the API driver will say
//                something like Device: ID: USB <-> Serial
//
//= OUTPUT VARIABLES
//
//- usbDevice        Pointer to a handle to modify.  On output it will
//                   contain the handle to use to talk to the TBAN.
//
//= RETURN VALUE
//
//- Returns NULL if everything worked.  Otherwise returns an error message.
//
SAMURIZE_PLUGIN_CHAR 
openConnection( int deviceId,
                FT_HANDLE* usbDevice )
{
   FT_STATUS status = FT_Open( deviceId, usbDevice );
   if ( status != FT_OK )
   {
      return "FT_OpenEx: Can't open requested device.";
   }

   status = FT_SetBaudRate( *usbDevice, FT_BAUD_19200 );
   if ( status != FT_OK )
   {
      FT_Close( *usbDevice );
      *usbDevice = 0;
      return "FT_SetBaudRate: Cant't set baud to 19200.";
   }
   
   status = FT_SetDataCharacteristics( *usbDevice, FT_BITS_8, FT_STOP_BITS_1,
                                       FT_PARITY_NONE );
   if ( status != FT_OK )
   {
      FT_Close( *usbDevice );
      *usbDevice = 0;
      return "FT_SetDataCharacteristics: Can't set 8 bit, 1 stop, no parity.";
   }

   // set a 3 second timeout
   status = FT_SetTimeouts( *usbDevice, 3000, 0 );
   if ( status != FT_OK )
   {
      FT_Close( *usbDevice );
      *usbDevice = 0;
      return "FT_SetTimeouts: Can't set a 3 second timeout.";
   }
   
   return 0;
}

//: Read the data from the TBAN.
//
//= INPUT VARIABLES
//
//- usbDevice     The handle to use to talk to the TBAN.
//
//= OUTPUT VARIABLES
//
//- data          Should be an array of 285 characters.  On output
//                it will contain the read data.
//
//= RETURN VALUE
//
//- Returns NULL if everything worked.  Otherwise returns an error message.
//
SAMURIZE_PLUGIN_CHAR
readFromTBan( FT_HANDLE usbDevice,
              char* data )
{
   // SER_Request command code. 
   char sendBuf[1];
   sendBuf[0] = 0x36;

   DWORD numWritten;
   FT_STATUS status = FT_Write( usbDevice, &sendBuf[0], 1, &numWritten);
   if ( status != FT_OK || numWritten != 1 )
   {
      return "FT_Write: Can't send SER_Request command.";
   }

   DWORD numRead;
   status = FT_Read( usbDevice, data, DATA_LEN, &numRead );
   if ( status != FT_OK )
   {
      // fatal error like disconnected USB port
      return "FT_Read: Can't read data from TBAN.";
   }
   else if ( numRead != DATA_LEN )
   {
      // probably a time out
      return "FT_Read: Didn't get all the data.";
   }

   // Update the maximum fan RPM values.
   computeMaxFanRPM( data );
   return 0;
}

//===========================================================================
//
// PRIVATE UTILITY FUNCTIONS
//
//===========================================================================

//: Compute the maximum fan RPM values from the data array.
//
// The equation can be found at:
// http://www.mcubed-tech.com/eng/support_kit.htm
//
//= INPUT VARIABLES
//
//- data          Should be an array of 285 characters.  The TBAN
//                data array to use.
//
static void
computeMaxFanRPM( char* data )
{
   // Compute the maximum fan RPM value.
   for ( int i = 0; i < 4; i++ )
   {
      // Max RPM is a 2 byte quantity starting at ID 149.  From what I
      // can tell, the protocol docs are wrong.  They list the high
      // byte first and the low byte second but experimentation proves
      // the reverse is true.
      int fanOffset = 149 - 1 + 2 * i;
      int highByte = (int)data[ fanOffset+1 ];
      int lowByte = (int)data[ fanOffset ];

      // From the T-Balancer support page: FanMaxSpeed: add the low- and
      // high byte to a 16bit value (256*HB+LB), multiply it with 10.5 to
      // receive the value for turns per minute.
      s_maxFanRPM[i] = (int)( 10.5 * ( 256 * highByte + lowByte ) );
   }
}

//: Fill the TBAN sensors and fan data with random values.
//
//= OUTPUT VARIABLES
//
//- data          Should be an array of 285 characters.  The sensor
//                and fan values will be entered.
//
static char*
readRandom( char* data )
{
   // Fill in digital temp values in the range 20-60.
   int base = 238;
   for ( int i = 0; i < 8; i++ )
   {
      double raw = (double)rand() / (double)RAND_MAX;
      int value = 20 + (int)( raw * ( 60 - 20 ) );
      s_data[base+i] = (char)( value );
   }

   // Fill in analog temp values in the range 20-60.
   base = 246;
   for ( i = 0; i < 6; i++ )
   {
      double raw = (double)rand() / (double)RAND_MAX;
      int value = 20 + (int)( raw * ( 60 - 20 ) );
      s_data[base+i] = (char)( value );
   }

   // Fill in Fan percentage values in the range 0-100.
   base = 137;
   for ( i = 0; i < 4; i++ )
   {
      double raw = (double)rand() / (double)RAND_MAX;
      int value = 0 + (int)( raw * ( 100 - 0 ) );
      s_data[base+i] = (char)( value );
   }

   // Fill in Max Fan RPM values in the range 1500-2000.
   base = 148;
   for ( i = 0; i < 8; i += 2 )
   {
      double raw = (double)rand() / (double)RAND_MAX;
      double rpm = 1500 + ( raw * ( 2000 - 1500 ) );

      int highByte = (int)floor( rpm / 256.0 );
      int lowByte = (int)( rpm - 256 * highByte );
      
      s_data[base+i] = (char)highByte;
      s_data[base+i+1] = (char)lowByte;
   }

   computeMaxFanRPM( s_data );
   return 0;
}

//: Dialog message handler
//
//= INPUT VARIABLES
//
//- hDialog    Handle to the dialog box. 
//- uMsg       Specifies the message. 
//- wParam     Specifies additional message-specific information. 
//- lParam     Specifies additional message-specific information. 
//
//= RETURN VALUE
//
//- Typically, the dialog box procedure should return TRUE if it
//  processed the message, and FALSE if it did not. If the dialog box
//  procedure returns FALSE, the dialog manager performs the default
//  dialog operation in response to the message.
//
static int CALLBACK
dialogCallback( HWND hDialog,
                UINT uMsg,
                WPARAM wParam,
                LPARAM lParam )
{
   switch( uMsg )
	{
      // Initialize the dialog.
   case WM_INITDIALOG:
      {
         // Load the existing settings.  Ignore any errors since they
         // may be the first time we make the settings.
         int refreshSec = 5;
         loadSettings( 0, s_deviceId, refreshSec );

         // Initialize the device ID
         char buffer[40];
         sprintf( buffer, "%d", s_deviceId );
         SetDlgItemText( hDialog, IDC_DEVICE_ID, buffer );
         
         // Initialize the refresh rate
         sprintf( buffer, "%d", refreshSec );
         SetDlgItemText( hDialog, IDC_REFRESH, buffer );

         return true;
      }
      break;

   case WM_COMMAND:
      switch ( LOWORD( wParam ) )
      {
      case IDOK:
         {
            // Get the device ID.  
            BOOL success = false;
            s_deviceId = GetDlgItemInt( hDialog, IDC_DEVICE_ID, &success,
                                        false);
            if ( ! success )
            {
               errorDialog( s_samurizeWindow, "Read devise ID" );
               return true;
            }
            
            int refreshSec = GetDlgItemInt( hDialog, IDC_REFRESH, &success,
                                            false);
            if ( ! success )
            {
               errorDialog( s_samurizeWindow, "Read refresh rate" );
               return true;
            }
            else if ( refreshSec <= 0 )
            {
               MessageBox( hDialog, "The refresh rate must be > 0.",
                           "Error", MB_OK );
               return true;
            }

            // Save the settings to the registry.
            int error = saveSettings( hDialog, s_deviceId, refreshSec );
            if ( error )
            {
               return true;
            }
         }
         // fall through
         
      case IDCANCEL:
         EndDialog( hDialog, true );
         return true;
      }

   default:
      return false;
   }

   return false;
}

//: Display the last error message.
//
// This function will take the windows error code and convert it to a
// human readable message and then show that in a dialog box.
//
//= INPUT VARIABLES
//
//- parent     The parent widget to use.
//- funcName   The name of the function that had the error.
//- errorCode  The error code to interpret.  If set to zero, GetLastError()
//             is used to get the code.
//
static void
errorDialog( HWND parent,
             const char* funcName,
             int errorCode )
{
   // Code stolen, cleaned up, and reformatted from here:
   // http://msdn.microsoft.com/library/default.asp?url=/library/en-us/debug/base/retrieving_the_last_error_code.asp
   if ( ! errorCode )
   {
      errorCode = GetLastError();
   }
   
   char* msgBuffer;
   FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER | 
                  FORMAT_MESSAGE_FROM_SYSTEM,
                  NULL,
                  errorCode,
                  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                  (LPTSTR) &msgBuffer,
                  0, NULL );

   char* displayBuffer = (char*)LocalAlloc( LMEM_ZEROINIT, 
                                            strlen( msgBuffer )+
                                            strlen( funcName ) + 40 ); 
   wsprintf( displayBuffer, "%s failed with error %d\n%s", 
             funcName, errorCode, msgBuffer ); 
   MessageBox( parent, displayBuffer, "Error", MB_OK ); 

   LocalFree( msgBuffer );
   LocalFree( displayBuffer );
}

//: Load the device ID and refresh rate from the registry
//
//= INPUT VARIABLES
//
//- parent        Handle to a parent widget.  Used to display error messages.
//                If this is set to null, no error dialogs will be shown.
//
//= OUTPUT VARIABLES
//
//- deviceId      The USB device ID number of the TBAN.
//- refreshSec    The refresh rate in seconds to save.
//
//= RETURN VALUE
//
//- Returns zero if successful.  Returns 1 otherwise.  An error dialog
//  is displayed to show any error messages.
//
static int
loadSettings( HWND parent,
              int& deviceId,
              int& refreshSec )
{
   // Open the key to the plugin data.
   HKEY key;
   int error = RegOpenKeyEx( HKEY_CURRENT_USER, s_tbanRegKey,
                             0,                // options
                             KEY_QUERY_VALUE,  // action
                             &key );     // result
   if ( error )
   {
      s_readerError = "Can't open registry key.";
      if ( parent ) errorDialog( parent, "RegOpenKeyEx", error );
      return 1;
   }

   // Read the device ID
   int numRead = sizeof( deviceId );
   error = RegQueryValueEx( key,
                            "Device ID", // key name
                            0, // flags
                            0, // output type code
                            (LPBYTE)&deviceId,
                            (LPDWORD)&numRead );
   if ( error || numRead != sizeof( deviceId ) )
   {
      RegCloseKey( key );
      s_readerError = "Can't read device ID.";
      if ( parent ) errorDialog( parent, "RegQueryValueEx", error );
      return 1;
   }

   // Read the refresh rate
   numRead = sizeof( refreshSec );
   error = RegQueryValueEx( key,
                            "Refresh Rate", // key name
                            0, // flags
                            0, // output type code
                            (LPBYTE)&refreshSec,
                            (LPDWORD)&numRead );
   if ( error || numRead != sizeof( refreshSec ) )
   {
      RegCloseKey( key );
      s_readerError = "Can't read refresh rate.";
      if ( parent ) errorDialog( parent, "RegQueryValueEx", error );
      return 1;
   }

   RegCloseKey( key );
   return 0;
}

//: Save the device ID and refresh rate to the registry
//
//= INPUT VARIABLES
//
//- parent        Handle to a parent widget.  Used to display error messages.
//                If this is set to null, no error dialogs will be shown.
//- deviceId      The USB device ID number of the TBAN.
//- refreshSec    The refresh rate in seconds to save.
//
//= RETURN VALUE
//
//- Returns zero if successful.  Returns 1 otherwise.  An error dialog
//  is displayed to show any error messages.
//
static int
saveSettings( HWND parent,
              int deviceId,
              int refreshSec )
{
   // Create the registry key.  This will open it if the key already exists.
   HKEY key;
   int error = RegCreateKeyEx( HKEY_CURRENT_USER, s_tbanRegKey,
                               0,      // reserved must be 0
                               0,      // key object type
                               REG_OPTION_NON_VOLATILE, // save the key
                               KEY_SET_VALUE,   // access rights
                               0,      // default security descriptor
                               &key,   // the returned key
                               0 );    // was it new or existing?
   if ( error )
   {
      s_readerError = "Can't create registry key.";
      if ( parent ) errorDialog( parent, "RegOpenKeyEx", error );
      return 1;
   }

   // Save the USB device ID.  
   int len = sizeof( deviceId );
   error = RegSetValueEx( key,
                          "Device ID",    // the value name
                          0,              // reserved must be 0
                          REG_DWORD,       // type code
                          (const BYTE*)&deviceId,
                          len );
   if ( error )
   {
      RegCloseKey( key );
      s_readerError = "Can't write device ID.";
      if ( parent ) errorDialog( parent, "RegSetValueEx", error );
      return 1;
   }

   // Save the refresh rate.
   len = sizeof( refreshSec );
   error = RegSetValueEx( key,
                          "Refresh Rate", // the value name
                          0,               // reserved must be 0
                          REG_DWORD,       // type code
                          (const BYTE*)&refreshSec,
                          len );
   if ( error )
   {
      RegCloseKey( key );
      s_readerError = "Can't write refresh rate.";
      if ( parent ) errorDialog( parent, "RegSetValueEx", error );
      return 1;
   }

   RegCloseKey( key );
   return 0;
}
